# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Onedata event

{"Key": "/my-onedata-space/files/file.txt",
 "Records": [{"objectKey": "file.txt",
              "objectId": "0000034500046EE9C6775...",
              "eventTime": "2019-02-07T09:51:04.347823",
              "eventSource": "OneTrigger"}]
}
"""

from faassupervisor.events.unknown import UnknownEvent


class OnedataEvent(UnknownEvent):
    """ Class to parse the Onedata event. """

    # pylint: disable=too-few-public-methods

    _TYPE = 'ONEDATA'

    def _set_event_params(self):
        self.object_key = self.event['Key']
        self.file_name = self.event_records['objectKey']



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This module provides a generic class
used to define storage providers."""

import abc


def get_bucket_name(output_path):
    """Returns the bucket name from a defined output path."""
    return output_path.split('/')[0]


def get_file_key(output_path, file_name):
    """Returns the correct 'file_key' required for uploading files."""
    stg_path = output_path.split('/', 1)
    # Path format => stg_path: bucket/<folder-path>
    # Last part is optional
    if len(stg_path) > 1:
        # There is a folder defined
        # Set the folder in the file path
        return f'{stg_path[1]}/{file_name}'
    return file_name


class DefaultStorageProvider(metaclass=abc.ABCMeta):
    """All the different data providers must inherit from this class
    to ensure that the commands are defined consistently."""

    _TYPE = 'DEFAULT'

    def __init__(self, stg_auth):
        self.stg_auth = stg_auth

    @abc.abstractmethod
    def download_file(self, parsed_event, input_dir_path):
        """Generic method to be implemented by all the storage providers."""

    @abc.abstractmethod
    def upload_file(self, file_path, file_name, output_path):
        """Generic method to be implemented by all the storage providers."""

    def get_type(self):
        """Returns the storage type.
        Can be LOCAL, MINIO, ONEDATA, S3."""
        return self._TYPE



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EventProvider identification flow:

       Create
      unknown
       event
         |
         V
    Is APIGateway? --(yes)--> Read Body, Parameters
         |                             |
        (no)                         (then)
         |-----------------------------|
         V
     Read EventProvider.
     Storage keys? --(yes)-->  Read keys (minio, s3, onedata)
         |                             |
        (no)                           |
         |-----------------------------|
         V
    Return parsed
       event
"""

import json
from faassupervisor.events.apigateway import ApiGatewayEvent
from faassupervisor.events.minio import MinioEvent
from faassupervisor.events.onedata import OnedataEvent
from faassupervisor.events.s3 import S3Event
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.logger import get_logger
from faassupervisor.exceptions import exception, UnknowStorageEventWarning
from faassupervisor.utils import SysUtils

_S3_EVENT = "aws:s3"
_MINIO_EVENT = "minio:s3"
_ONEDATA_EVENT = "OneTrigger"


def _is_api_gateway_event(event_info):
    return 'httpMethod' in event_info


def _is_storage_event(event_info):
    if 'Records' in event_info \
           and event_info['Records'] \
           and 'eventSource' in event_info['Records'][0]:
        return event_info['Records'][0]['eventSource'] == _S3_EVENT \
            or event_info['Records'][0]['eventSource'] == _MINIO_EVENT \
            or event_info['Records'][0]['eventSource'] == _ONEDATA_EVENT
    return False


@exception()
def _parse_storage_event(event):
    record = event['Records'][0]['eventSource']
    if record == _S3_EVENT:
        parsed_event = S3Event(event)
        get_logger().info("S3 event created")
    elif record == _MINIO_EVENT:
        parsed_event = MinioEvent(event)
        get_logger().info("MINIO event created")
    elif record == _ONEDATA_EVENT:
        parsed_event = OnedataEvent(event)
        get_logger().info("ONEDATA event created")
    else:
        raise UnknowStorageEventWarning()
    return parsed_event


def parse_event(event):
    """Parses the received event and
    returns the appropriate event class."""
    # Make sure the event is always stored
    parsed_event = None
    if not isinstance(event, dict):
        try:
            event = json.loads(event)
        except ValueError:
            return UnknownEvent(event)
    # Applies the event identification flow
    if _is_api_gateway_event(event):
        get_logger().info("API Gateway event found.")
        parsed_event = ApiGatewayEvent(event)
        # Update event info with API request event body
        # to be further processed (if needed)
        if parsed_event.has_json_body():
            event = parsed_event.body
            if not isinstance(parsed_event.body, dict):
                event = json.loads(parsed_event.body)
    if _is_storage_event(event):
        get_logger().info("Storage event found.")
        parsed_event = _parse_storage_event(event)
        # Store 'object_key' in environment variable
        SysUtils.set_env_var("STORAGE_OBJECT_KEY", parsed_event.object_key)
    return parsed_event if parsed_event else UnknownEvent(event)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""S3 event exaple:

{'Records': [{'awsRegion': 'us-east-1',
              'eventName': 'ObjectCreated:Put',
              'eventSource': 'aws:s3',
              'eventTime': '2019-02-23T11:40:46.473Z',
              'eventVersion': '2.1',
              'requestParameters': {'sourceIPAddress': '84.123.4.23'},
              'responseElements': {'x-amz-id-2': 'XXXXX',
                                   'x-amz-request-id': 'XXXXX'},
              's3': {'bucket': {'arn': 'arn:aws:s3:::scar-darknet-bucket',
                                'name': 'scar-darknet-bucket',
                                'ownerIdentity': {'principalId': 'XXXXX'}},
                     'configurationId': 'XXXXX',
                     'object': {'eTag': 'XXXXX',
                                'key': 'scar-darknet-s3/input/dog.jpg',
                                'sequencer': 'XXXXX',
                                'size': 999},
                     's3SchemaVersion': '1.0'},
              'userIdentity': {'principalId': 'AWS:XXXXX'}}]}

"""
from urllib.parse import unquote_plus
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.utils import FileUtils


class S3Event(UnknownEvent):
    """ Class to parse the S3 event. """

    # pylint: disable=too-few-public-methods

    _TYPE = 'S3'

    def _set_event_params(self):
        self.bucket_arn = self.event_records['s3']['bucket']['arn']
        self.bucket_name = self.event_records['s3']['bucket']['name']
        self.object_key = unquote_plus(self.event_records['s3']['object']['key'])
        self.file_name = FileUtils.get_file_name(self.object_key)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with all the generic supervisor classes and methods.
Also entry point of the faassupervisor package."""

from faassupervisor.events import parse_event
from faassupervisor.exceptions import exception, FaasSupervisorError
from faassupervisor.storage.config import StorageConfig
from faassupervisor.utils import SysUtils, FileUtils
from faassupervisor.logger import configure_logger, get_logger
from faassupervisor.faas.aws_lambda.supervisor import LambdaSupervisor, is_batch_execution
from faassupervisor.faas.binary.supervisor import BinarySupervisor


class Supervisor():
    """Generic supervisor used to create the required supervisors
    based on the environment variable 'SUPERVISOR_TYPE'."""

    # pylint: disable=too-few-public-methods

    def __init__(self, event, context=None):
        self._create_tmp_dirs()
        # Parse the event_info data
        self.parsed_event = parse_event(event)
        # Read storage config
        self._read_storage_config()
        # Create the supervisor
        self.supervisor = _create_supervisor(event, context)

    def _create_tmp_dirs(self):
        """Creates the temporal directories where the
        input/output data is going to be stored.

        The folders are deleted automatically
        when the execution finishes.
        """
        self.input_tmp_dir = FileUtils.create_tmp_dir()
        self.output_tmp_dir = FileUtils.create_tmp_dir()
        SysUtils.set_env_var("TMP_INPUT_DIR", self.input_tmp_dir.name)
        SysUtils.set_env_var("TMP_OUTPUT_DIR", self.output_tmp_dir.name)

    def _read_storage_config(self):
        get_logger().info("Reading storage configuration")
        self.stg_config = StorageConfig()

    @exception()
    def _parse_input(self):
        """Download input data from storage provider
        or save data from POST request.

        A function can have information from several storage providers
        but one event always represents only one file (so far), so only
        one provider is going to be used for each event received.
        """
        input_file_path = self.stg_config.download_input(self.parsed_event,
                                                         self.input_tmp_dir.name)
        if input_file_path and FileUtils.is_file(input_file_path):
            SysUtils.set_env_var('INPUT_FILE_PATH', input_file_path)
            get_logger().info('INPUT_FILE_PATH variable set to \'%s\'', input_file_path)

    @exception()
    def _parse_output(self):
        self.stg_config.upload_output(self.output_tmp_dir.name)

    @exception()
    def run(self):
        """Generic method to launch the supervisor execution."""
        try:
            if is_batch_execution() and SysUtils.is_lambda_environment():
                # Only delegate to batch
                self.supervisor.execute_function()
            else:
                self._parse_input()
                self.supervisor.execute_function()
                self._parse_output()
            get_logger().info('Creating response')
            return self.supervisor.create_response()
        except FaasSupervisorError as fse:
            get_logger().exception(fse)
            get_logger().error('Creating error response')
            return self.supervisor.create_error_response()


@exception()
def _create_supervisor(event, context=None):
    """Returns a new supervisor based on the
    environment.
    Binary mode by default"""
    supervisor = None
    if SysUtils.is_lambda_environment():
        supervisor = LambdaSupervisor(event, context)
    else:
        supervisor = BinarySupervisor()
    return supervisor


def main(event, context=None):
    """Initializes the generic supervisor
    and launches its execution."""
    configure_logger()
    get_logger().debug("EVENT received: %s", event)
    if context:
        get_logger().debug("CONTEXT received: %s", context)
    supervisor = Supervisor(event, context)
    return supervisor.run()


if __name__ == "__main__":
    # If supervisor is running as a binary
    # receive the input from stdin.
    ret = main(SysUtils.get_stdin())
    if ret is not None:
        print(ret)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the custom exceptions for the supervisor. """

import functools
import sys
from requests.exceptions import RequestException
from botocore.exceptions import ClientError
from faassupervisor.logger import get_logger


def exception():
    """A decorator that wraps the passed in function and logs exceptions."""

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)

            except ClientError as cerr:
                print(f"There was an exception in {func.__name__}")
                print(cerr.response['Error']['Message'])
                get_logger().error(cerr)
                sys.exit(1)

            except RequestException as rexc:
                print(f"There was an exception in {func.__name__}")
                get_logger().error(rexc)
                sys.exit(1)

            except FaasSupervisorError as fse:
                # print(fse.args[0])
                # get_logger().error(fse)
                if 'Warning' in fse.__class__.__name__:
                    get_logger().warning(fse)
                # Finish the execution if it's an error
                if 'Error' in fse.__class__.__name__:
                    get_logger().error(fse)
                    sys.exit(1)

        return wrapper

    return decorator


class FaasSupervisorError(Exception):
    """
    The base exception class for exceptions.

    :ivar msg: The descriptive message associated with the error.
    """
    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


################################################
##             GENERAL EXCEPTIONS             ##
################################################
class InvalidPlatformError(FaasSupervisorError):
    """
    The binary is not launched on a Linux platform

    """
    fmt = "This binary only works on a Linux Platform.\nTry executing the Python version."


class InvalidSupervisorTypeError(FaasSupervisorError):
    """
    The supervisor type is not in the allowed list

    """
    fmt = "The supervisor type '{sup_typ}' is not allowed."


class InvalidStoragePathTypeError(FaasSupervisorError):
    """
    The storage path type is not in the allowed list

    """
    fmt = "The storage path type '{storage_type}' is not allowed."


class ContainerImageNotFoundError(FaasSupervisorError):
    """
    The container image is no specified

    """
    fmt = "Container image id is not specified."


class ContainerTimeoutExpiredWarning(FaasSupervisorError):
    """
    The udocker containers has exceeded the defined execution time.

    """
    fmt = "Container timeout expired.\nContainer execution stopped."


class NoLambdaContextError(FaasSupervisorError):
    """
    No context was provided for the lambda instance.

    """
    fmt = "No context found in the Lambda environment."


class UnknowStorageEventWarning(FaasSupervisorError):
    """
    Unknown storage event detected

    """
    fmt = "Unknown storage event detected."


################################################
##        STORAGE PROVIDER EXCEPTIONS         ##
################################################
class InvalidStorageProviderError(FaasSupervisorError):
    """
    The storage provider type is not valid.

    """
    fmt = "Invalid storage provider type defined: '{storage_type}'."


class NoStorageProviderDefinedWarning(FaasSupervisorError):
    """
    There is no storage provider defined.

    """
    fmt = "There is no storage provider defined for this function execution."


class StorageTypeError(FaasSupervisorError):
    """
    The storage type defined is not allowed

    """
    fmt = "The storage type '{auth_type}' is not allowed."


class StorageAuthError(FaasSupervisorError):
    """
    The storage authentication is not well-defined.

    """
    fmt = "The storage authentication of '{auth_type}' is not well-defined."


################################################
##        ONEDATA PROVIDER EXCEPTIONS         ##
################################################
class OnedataUploadError(FaasSupervisorError):
    """
    Uploading file to Onedata failed.
    """
    fmt = ("Uploading file '{file_name}' to Onedata failed. "
           "Status code: {status_code}")


class OnedataDownloadError(FaasSupervisorError):
    """
    Downloading file from Onedata failed.
    """
    fmt = ("Downloading file '{file_name}' from Onedata failed. "
           "Status code: {status_code}")


class OnedataFolderCreationError(FaasSupervisorError):
    """
    Folder creation in Onedata failed.
    """
    fmt = ("Folder '{folder_name}' creation in Onedata failed. "
           "Status code: {status_code}")



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In this module are defined all the methods and classes used
to manage a udocker container in the lambda environment."""

import subprocess
from faassupervisor.exceptions import ContainerImageNotFoundError
from faassupervisor.utils import SysUtils, FileUtils, ConfigUtils
from faassupervisor.logger import get_logger
from faassupervisor.exceptions import ContainerTimeoutExpiredWarning
from faassupervisor.faas.aws_lambda.function import get_function_ip


def _parse_cont_env_var(key, value):
    return ["--env", str(key) + '=' + str(value)] if key and value else []


class Udocker():
    """Class in charge of managing the udocker binary."""

    _CONTAINER_OUTPUT_FILE = SysUtils.join_paths(FileUtils.get_tmp_dir(), "container-stdout")
    _CONTAINER_NAME = "udocker_container"
    _SCRIPT_EXEC = "/bin/sh"

    def __init__(self, lambda_instance):
        self.lambda_instance = lambda_instance
        # Create required udocker folder
        FileUtils.create_folder(SysUtils.get_env_var("UDOCKER_DIR"))
        # Init the udocker command that will be executed
        self.udocker_exec = [SysUtils.get_env_var("UDOCKER_EXEC")]
        self.cont_cmd = self.udocker_exec + ["--quiet", "run"]

        self.cont_img_id = ConfigUtils.read_cfg_var('container').get('image')
        if not self.cont_img_id:
            raise ContainerImageNotFoundError()

    def _list_udocker_images_cmd(self):
        return self.udocker_exec + ["images"]

    def _load_udocker_image_cmd(self):
        return self.udocker_exec + ["load", "-i", self.cont_img_id]

    def _download_udocker_image_cmd(self):
        return self.udocker_exec + ["pull", self.cont_img_id]

    def _list_udocker_containers_cmd(self):
        return self.udocker_exec + ["ps"]

    def _create_udocker_container_cmd(self):
        return self.udocker_exec + ["create", f"--name={self._CONTAINER_NAME}", self.cont_img_id]

    def _set_udocker_container_execution_mode_cmd(self):
        return self.udocker_exec + ["setup", "--execmode=F1", self._CONTAINER_NAME]

    def _is_container_image_downloaded(self):
        cmd_out = SysUtils.execute_cmd_and_return_output(self._list_udocker_images_cmd())
        return self.cont_img_id in cmd_out

    def _load_local_container_image(self):
        get_logger().info("Loading container image '%s'", self.cont_img_id)
        SysUtils.execute_cmd(self._load_udocker_image_cmd())

    def _download_container_image(self):
        get_logger().info("Pulling container '%s' from Docker Hub", self.cont_img_id)
        SysUtils.execute_cmd(self._download_udocker_image_cmd())

    def _is_container_available(self):
        cmd_out = SysUtils.execute_cmd_and_return_output(self._list_udocker_containers_cmd())
        return self._CONTAINER_NAME in cmd_out

    def _create_image(self):
        if self._is_container_image_downloaded():
            get_logger().info("Container image '%s' already available", self.cont_img_id)
        else:
            if SysUtils.is_var_in_env("IMAGE_FILE"):
                self._load_local_container_image()
            else:
                self._download_container_image()

    def _create_container(self):
        if self._is_container_available():
            get_logger().info("Container already available")
        else:
            get_logger().info("Creating container based on image '%s'.", self.cont_img_id)
            SysUtils.execute_cmd(self._create_udocker_container_cmd())
        SysUtils.execute_cmd(self._set_udocker_container_execution_mode_cmd())

    def _create_command(self):
        self._add_container_volumes()
        self._add_container_environment_variables()
        # Container running script
        if hasattr(self.lambda_instance, 'script_path'):
            # Add script in memory as entrypoint
            self.cont_cmd += [(f"--entrypoint={self._SCRIPT_EXEC} "
                               f"{self.lambda_instance.script_path}"),
                              self._CONTAINER_NAME]
        # Container with args
        elif hasattr(self.lambda_instance, 'cmd_args'):
            # Add args
            self.cont_cmd += [self._CONTAINER_NAME]
            self.cont_cmd += self.lambda_instance.cmd_args
        # Script to be executed every time (if defined)
        elif hasattr(self.lambda_instance, 'init_script_path'):
            # Add init script
            self.cont_cmd += [(f"--entrypoint={self._SCRIPT_EXEC} "
                               f"{self.lambda_instance.init_script_path}"),
                              self._CONTAINER_NAME]
        # Only container
        else:
            self.cont_cmd += [self._CONTAINER_NAME]

    def _add_container_volumes(self):
        self.cont_cmd.extend(["-v", SysUtils.get_env_var("TMP_INPUT_DIR")])
        self.cont_cmd.extend(["-v", SysUtils.get_env_var("TMP_OUTPUT_DIR")])
        self.cont_cmd.extend(["-v", "/dev", "-v", "/proc", "-v", "/etc/hosts", "--nosysdirs"])
        if SysUtils.is_var_in_env('EXTRA_PAYLOAD'):
            self.cont_cmd.extend(["-v", self.lambda_instance.PERMANENT_FOLDER])

    def _add_cont_env_vars(self):
        for key, value in SysUtils.get_cont_env_vars().items():
            self.cont_cmd.extend(_parse_cont_env_var(key, value))

    def _add_input_file(self):
        self.cont_cmd.extend(_parse_cont_env_var("INPUT_FILE_PATH",
                                                 SysUtils.get_env_var("INPUT_FILE_PATH")))

    def _add_output_dir(self):
        self.cont_cmd.extend(_parse_cont_env_var("TMP_OUTPUT_DIR",
                                                 SysUtils.get_env_var("TMP_OUTPUT_DIR")))

    def _add_storage_object_key(self):
        self.cont_cmd.extend(_parse_cont_env_var("STORAGE_OBJECT_KEY",
                                                 SysUtils.get_env_var("STORAGE_OBJECT_KEY")))

    def _add_extra_payload_path(self):
        self.cont_cmd.extend(_parse_cont_env_var("EXTRA_PAYLOAD",
                                                 SysUtils.get_env_var("EXTRA_PAYLOAD")))

    def _add_function_request_id(self):
        self.cont_cmd.extend(_parse_cont_env_var("REQUEST_ID",
                                                 self.lambda_instance.get_request_id()))

    def _add_aws_access_keys(self):
        self.cont_cmd.extend(_parse_cont_env_var("AWS_ACCESS_KEY_ID",
                                                 SysUtils.get_env_var("AWS_ACCESS_KEY_ID")))
        self.cont_cmd.extend(_parse_cont_env_var("AWS_SECRET_ACCESS_KEY",
                                                 SysUtils.get_env_var("AWS_SECRET_ACCESS_KEY")))
        self.cont_cmd.extend(_parse_cont_env_var("AWS_SESSION_TOKEN",
                                                 SysUtils.get_env_var("AWS_SESSION_TOKEN")))

    def _add_function_ip(self):
        self.cont_cmd.extend(_parse_cont_env_var("INSTANCE_IP", get_function_ip()))

    def _add_container_environment_variables(self):
        self._add_function_request_id()
        self._add_function_ip()
        self._add_aws_access_keys()
        self._add_cont_env_vars()
        self._add_input_file()
        self._add_output_dir()
        self._add_storage_object_key()
        self._add_extra_payload_path()

    def prepare_container(self):
        """Prepares the environment to execute the udocker container."""
        self._create_image()
        self._create_container()
        self._create_command()

    def launch_udocker_container(self):
        """Launches the udocker container.
        If the execution time of the container exceeds the defined execution time,
        the container is killed and a warning is raised."""
        remaining_seconds = self.lambda_instance.get_remaining_time_in_seconds()
        get_logger().info("Executing udocker container. Timeout set to '%d' seconds",
                          remaining_seconds)
        get_logger().debug("Udocker command: '%s'", self.cont_cmd)
        with open(self._CONTAINER_OUTPUT_FILE, "wb") as out:
            with subprocess.Popen(self.cont_cmd,
                                  stderr=subprocess.STDOUT,
                                  stdout=out,
                                  start_new_session=True) as process:
                try:
                    process.wait(timeout=remaining_seconds)
                except subprocess.TimeoutExpired:
                    get_logger().info("Stopping process '%s'", process)
                    process.kill()
                    raise ContainerTimeoutExpiredWarning()
        udocker_output = b''
        if FileUtils.is_file(self._CONTAINER_OUTPUT_FILE):
            udocker_output = FileUtils.read_file(self._CONTAINER_OUTPUT_FILE, file_mode="rb")
        return udocker_output



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the local storage provider. """

from faassupervisor.storage.providers import DefaultStorageProvider


class Local(DefaultStorageProvider):
    """Class to manage saving files in local storage."""

    _TYPE = 'LOCAL'

    def download_file(self, parsed_event, input_dir_path):
        """Delegates the 'download' and local storage to the event."""
        return parsed_event.save_event(input_dir_path)

    def upload_file(self, file_path, file_name, output_path):
        pass



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module with a generic supervisor definition."""

import abc


class DefaultSupervisor(metaclass=abc.ABCMeta):
    """All the different supervisors must inherit from this class
    to ensure that the commands are defined."""

    @abc.abstractmethod
    def execute_function(self):
        """Executes the function code
        (udocker container or user script)."""

    @abc.abstractmethod
    def create_response(self):
        """Creates the function response."""

    @abc.abstractmethod
    def create_error_response(self):
        """Creates an error response when something fails."""



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In this module are defined the classes and methods
used to manage the batch jobs from the lambda environment."""

import json
import boto3


class Batch():
    """Used for defining Batch jobs in the Batch environment."""

    # pylint: disable=too-few-public-methods

    def __init__(self, lambda_instance):
        self.lambda_instance = lambda_instance
        self.script = self._get_user_script()
        self._create_context()
        self._set_job_variables()
        self._create_batch_client()

    def _create_context(self):
        self.context = {
            'function_name': self.lambda_instance.get_function_name(),
            'memory_limit_in_mb': self.lambda_instance.get_memory(),
            'aws_request_id': self.lambda_instance.get_request_id(),
            'log_group_name': self.lambda_instance.get_log_group_name(),
            'log_stream_name': self.lambda_instance.get_log_stream_name()
        }

    def _create_batch_client(self):
        self.client = boto3.client('batch')

    def _set_job_variables(self):
        self.batch_job_env_vars = []
        if self.script:
            self._add_batch_job_env_var("SCRIPT", self.script)
        self._add_batch_job_env_var("EVENT", json.dumps(self.lambda_instance.raw_event))
        self._add_batch_job_env_var("CONTEXT", json.dumps(self.context))
        self._add_batch_job_env_var("AWS_LAMBDA_REQUEST_ID", self.context.get("aws_request_id"))

    def _add_batch_job_env_var(self, name, value):
        if name and value:
            self.batch_job_env_vars.append({"name": name, "value": value})

    def _get_user_script(self):
        script = ''
        if 'script' in self.lambda_instance.raw_event:
            script = self.lambda_instance.raw_event['script']
        return script

    def _get_job_args(self):
        job_name = self.context.get("function_name")
        job_def = {
            "jobDefinition": job_name,
            "jobName": job_name,
            "jobQueue": job_name,
            "containerOverrides": {
                "environment": self.batch_job_env_vars
            }
        }
        return job_def

    def _submit_batch_job(self, job_args):
        return self.client.submit_job(**job_args)["jobId"]

    def invoke_batch_function(self):
        """Submit job from a lambda instance."""
        return self._submit_batch_job(self._get_job_args())



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with all the classes and methods
related with the binary supervisor."""

import subprocess
import sys
from faassupervisor.faas import DefaultSupervisor
from faassupervisor.logger import get_logger
from faassupervisor.utils import SysUtils, FileUtils, StrUtils


class BinarySupervisor(DefaultSupervisor):
    """Supervisor class used in the Binary environment."""

    _SCRIPT_FILE_NAME = 'script.sh'
    _OSCAR_SCRIPT_PATH = '/oscar/config/script.sh'

    def __init__(self):
        self.output = ''
        get_logger().info('SUPERVISOR: Initializing Binary supervisor')

    def _get_script_path(self):
        script_path = None
        if SysUtils.is_var_in_env('SCRIPT'):
            script_path = SysUtils.join_paths(SysUtils.get_env_var("TMP_INPUT_DIR"),
                                              self._SCRIPT_FILE_NAME)
            script_content = StrUtils.base64_to_str(SysUtils.get_env_var('SCRIPT'))
            FileUtils.create_file_with_content(script_path, script_content)
            get_logger().info("Script file created in '%s'", script_path)
        elif FileUtils.is_file(self._OSCAR_SCRIPT_PATH):
            script_path = self._OSCAR_SCRIPT_PATH
            get_logger().info("Script file found in '%s'", script_path)
        return script_path

    def execute_function(self):
        script_path = self._get_script_path()
        if script_path:
            try:
                pyinstaller_library_path = SysUtils.get_env_var('LD_LIBRARY_PATH')
                orig_library_path = SysUtils.get_env_var('LD_LIBRARY_PATH_ORIG')
                if orig_library_path:
                    SysUtils.set_env_var('LD_LIBRARY_PATH', orig_library_path)
                self.output = subprocess.check_output(['/bin/sh', script_path],
                                                      stderr=subprocess.STDOUT).decode("latin-1")
                SysUtils.set_env_var('LD_LIBRARY_PATH', pyinstaller_library_path)
                get_logger().debug("CONTAINER OUTPUT:\n %s", self.output)
            except subprocess.CalledProcessError as cpe:
                # Exit with user script return code if an
                # error occurs (Kubernetes handles the error)
                get_logger().error(cpe.output.decode('latin-1'))
                sys.exit(cpe.returncode)
        else:
            get_logger().error('No user script found!')

    def create_response(self):
        return self.output

    def create_error_response(self):
        pass



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the Onedata storage provider. """

import requests
from faassupervisor.logger import get_logger
from faassupervisor.storage.providers import DefaultStorageProvider
from faassupervisor.utils import SysUtils, FileUtils
from faassupervisor.exceptions import OnedataDownloadError, \
    OnedataUploadError, OnedataFolderCreationError


class Onedata(DefaultStorageProvider):
    """Class that manages downloads and uploads from Onedata. """

    _TYPE = 'ONEDATA'
    _CDMI_PATH = '/cdmi'
    _CDMI_VERSION_HEADER = {'X-CDMI-Specification-Version': '1.1.1'}

    def __init__(self, stg_auth):
        super().__init__(stg_auth)
        self._set_onedata_environment()

    def _set_onedata_environment(self):
        self.oneprovider_space = self.stg_auth.get_credential('space')
        self.oneprovider_host = self.stg_auth.get_credential('oneprovider_host')
        self.headers = {'X-Auth-Token': self.stg_auth.get_credential('token')}

    def _create_folder(self, folder_name):
        url = (f'https://{self.oneprovider_host}{self._CDMI_PATH}/'
               f'{self.oneprovider_space}/{folder_name}/')
        response = requests.put(url, headers=self.headers)
        if response.status_code != 201:
            raise OnedataFolderCreationError(folder_name=folder_name,
                                             status_code=response.status_code)

    def _folder_exists(self, folder_name):
        url = (f'https://{self.oneprovider_host}{self._CDMI_PATH}/'
               f'{self.oneprovider_space}/{folder_name}/')
        headers = {**self._CDMI_VERSION_HEADER, **self.headers}
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            return True
        return False

    def download_file(self, parsed_event, input_dir_path):
        """Downloads the file from the space of Onedata and
        returns the path were the download is placed. """
        file_download_path = ""
        url = f'https://{self.oneprovider_host}{self._CDMI_PATH}{parsed_event.object_key}'
        get_logger().info('Downloading item from host \'%s\' with key \'%s\'',
                          self.oneprovider_host,
                          parsed_event.object_key)
        response = requests.get(url, headers=self.headers)
        if response.status_code == 200:
            file_download_path = SysUtils.join_paths(input_dir_path, parsed_event.file_name)
            FileUtils.create_file_with_content(file_download_path, response.content, mode='wb')

            get_logger().info('Successful download of file \'%s\' with key \'%s\' in path \'%s\'',
                              parsed_event.file_name,
                              parsed_event.object_key,
                              file_download_path)
        else:
            raise OnedataDownloadError(file_name=parsed_event.object_key,
                                       status_code=response.status_code)
        return file_download_path

    def upload_file(self, file_path, file_name, output_path):
        """Uploads the file to the Onedata output path."""
        file_name = file_name.strip('/')
        output_path = output_path.strip('/')
        upload_path = f'{output_path}/{file_name}'
        upload_folder = FileUtils.get_dir_name(upload_path)
        # Create output folder (and subfolders) if it does not exists
        if not self._folder_exists(upload_folder):
            folders = upload_folder.split('/')
            path = ''
            for folder in folders:
                path = f'{path}/{folder}'
                if not self._folder_exists(path):
                    self._create_folder(path)
        # Upload the file
        url = (f'https://{self.oneprovider_host}{self._CDMI_PATH}/'
               f'{self.oneprovider_space}/{upload_path}')
        get_logger().info('Uploading file \'%s\' to space \'%s\'',
                          upload_path,
                          self.oneprovider_space)
        with open(file_path, 'rb') as data:
            response = requests.put(url, data=data, headers=self.headers)
            if response.status_code not in [201, 202, 204]:
                raise OnedataUploadError(file_name=file_name,
                                         status_code=response.status_code)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the Minio storage provider. """

import boto3
from faassupervisor.storage.providers.s3 import S3


class Minio(S3):
    """Class that manages downloads and uploads from Minio. """

    _DEFAULT_MINIO_ENDPOINT = 'http://minio-service.minio:9000'
    _TYPE = 'MINIO'

    def _get_client(self):
        """Return Minio client with user configuration."""
        endpoint = self.stg_auth.get_credential('endpoint')
        if endpoint == '':
            endpoint = self._DEFAULT_MINIO_ENDPOINT
        verify = self.stg_auth.get_credential('verify')
        if verify == '':
            verify = True
        region = self.stg_auth.get_credential('region')
        if region == '':
            region = None
        return boto3.client('s3',
                            endpoint_url=endpoint,
                            region_name=region,
                            verify=verify,
                            aws_access_key_id=self.stg_auth.get_credential('access_key'),
                            aws_secret_access_key=self.stg_auth.get_credential('secret_key'))




# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module to define log configuration and management."""

import logging
from faassupervisor.utils import ConfigUtils


def _get_log_level():
    loglevel = logging.INFO
    config_level = ConfigUtils.read_cfg_var('log_level')
    if config_level is not '':
        loglevel = logging.getLevelName(config_level)
    return loglevel


def _get_stream_handler():
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    str_hdl = logging.StreamHandler()
    str_hdl.setFormatter(formatter)
    return str_hdl


def configure_logger():
    """Configure the global logger used by all the classes."""
    logger = logging.getLogger('supervisor')
    # Avoid initializing the logger several times
    if not logger.handlers:
        # Set logger configuration
        logger.propagate = 0
        logger.setLevel(_get_log_level())
        logger.addHandler(_get_stream_handler())


def get_logger():
    """Returns the configured logger."""
    return logging.getLogger('supervisor')



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with all the classes and methods
related with the AWS Lambda supervisor."""

import subprocess
import traceback
from faassupervisor.faas.aws_lambda.batch import Batch
from faassupervisor.faas.aws_lambda.function import LambdaInstance
from faassupervisor.faas.aws_lambda.udocker import Udocker
from faassupervisor.faas import DefaultSupervisor
from faassupervisor.logger import get_logger
from faassupervisor.utils import ConfigUtils, StrUtils
from faassupervisor.exceptions import NoLambdaContextError, \
    ContainerTimeoutExpiredWarning


def is_batch_execution():
    return ConfigUtils.read_cfg_var("execution_mode") == "batch"


def _is_lambda_batch_execution():
    return ConfigUtils.read_cfg_var("execution_mode") == "lambda-batch"


class LambdaSupervisor(DefaultSupervisor):
    """Supervisor class used in the Lambda environment."""

    def __init__(self, event, context):
        if context:
            get_logger().info('SUPERVISOR: Initializing AWS Lambda supervisor')
            self.lambda_instance = LambdaInstance(event, context)
            self.body = {}
        else:
            raise NoLambdaContextError()

    def _execute_batch(self):
        batch_ri = Batch(self.lambda_instance).invoke_batch_function()
        batch_logs = (f"Job delegated to batch.\n"
                      f"Check batch logs with:\n"
                      f"  scar log -n {self.lambda_instance.get_function_name()} -ri {batch_ri}")
        get_logger().info(batch_logs)
        self.body["udocker_output"] = batch_logs.encode('utf-8')

    def _execute_udocker(self):
        try:
            udocker = Udocker(self.lambda_instance)
            udocker.prepare_container()
            self.body["udocker_output"] = udocker.launch_udocker_container()
            get_logger().debug("CONTAINER OUTPUT:\n %s", self.body["udocker_output"].decode("latin-1"))
        except (subprocess.TimeoutExpired, ContainerTimeoutExpiredWarning):
            get_logger().warning("Container execution timed out")
            if _is_lambda_batch_execution():
                self._execute_batch()

    def execute_function(self):
        if is_batch_execution():
            self._execute_batch()
        else:
            self._execute_udocker()

    def create_error_response(self):
        exception_msg = traceback.format_exc()
        get_logger().error("Exception launched:\n %s", exception_msg)
        return {
            "statusCode": 500,
            "headers": {
                "amz-lambda-request-id": self.lambda_instance.get_request_id(),
                "amz-log-group-name": self.lambda_instance.get_log_group_name(),
                "amz-log-stream-name": self.lambda_instance.get_log_stream_name()
            },
            "body": StrUtils.dict_to_base64str({"exception": exception_msg}),
            "isBase64Encoded": True,
        }

    def create_response(self):
        return {
            "statusCode": 200,
            "headers": {
                "amz-lambda-request-id": self.lambda_instance.get_request_id(),
                "amz-log-group-name": self.lambda_instance.get_log_group_name(),
                "amz-log-stream-name": self.lambda_instance.get_log_stream_name()
            },
            "body": StrUtils.bytes_to_base64str(self.body["udocker_output"]),
            "isBase64Encoded": True,
        }



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods shared by all the classes."""

import base64
import json
import os
import subprocess
import sys
import tempfile
import shutil
import yaml


class SysUtils():
    """Common methods for system management."""

    @staticmethod
    def get_stdin():
        """Returns system stdin."""
        return sys.stdin.read()

    @staticmethod
    def join_paths(*paths):
        """Returns the strings passed joined as a system path."""
        return os.path.join(*paths)

    @staticmethod
    def is_var_in_env(variable):
        """Checks if a variable is in the system environment."""
        return variable in os.environ

    @staticmethod
    def set_env_var(key, value):
        """Sets a system environment variable."""
        if key and value:
            os.environ[key] = value

    @staticmethod
    def get_all_env_vars():
        """Returns all the system environment variables."""
        return os.environ

    @staticmethod
    def get_env_var(variable):
        """Returns the value of system environment variable
        or an empty string if not found."""
        return os.environ.get(variable, "")

    @staticmethod
    def get_cont_env_vars():
        """Returns the defined container environment variables."""
        container = ConfigUtils.read_cfg_var('container')
        return container.get('environment', {}).get('Variables', {})

    @staticmethod
    def execute_cmd(command):
        """Executes a bash command."""
        subprocess.call(command)

    @staticmethod
    def execute_cmd_and_return_output(command, encoding='utf-8'):
        """Executes a bash command and returns the console output."""
        return subprocess.check_output(command).decode(encoding)

    @staticmethod
    def is_lambda_environment():
        """Checks if supervisor is running in AWS Lambda."""
        return (SysUtils.is_var_in_env('AWS_EXECUTION_ENV') and
                SysUtils.get_env_var('AWS_EXECUTION_ENV').startswith('AWS_Lambda_'))


class FileUtils():
    """Common methods for file and directory management."""

    @staticmethod
    def set_file_execution_rights(file_path):
        """Makes a file executable."""
        # Execution rights for user, group and others
        mode = os.stat(file_path).st_mode | 0o0111
        os.chmod(file_path, mode)

    @staticmethod
    def cp_file(file_src, file_dst):
        """Copy file to specified destination."""
        shutil.copyfile(file_src, file_dst)

    @staticmethod
    def create_folder(folder_name):
        """Creates a system folder.
        Does nothing if the folder exists."""
        os.makedirs(folder_name, exist_ok=True)

    @staticmethod
    def create_file_with_content(path, content, mode='w'):
        """Creates a new file with the passed content.
        If the content is a dictionary, first is converted to a string."""
        with open(path, mode) as fwc:
            if isinstance(content, dict):
                content = json.dumps(content)
            fwc.write(content)

    @staticmethod
    def read_file(file_path, file_mode="r", file_encoding="utf-8"):
        """Reads the whole specified file and returns the content."""
        content = ''
        if file_mode == 'rb':
            file_encoding = None
        with open(file_path, mode=file_mode, encoding=file_encoding) as content_file:
            content = content_file.read()
        return content

    @staticmethod
    def create_tmp_dir():
        """Creates a directory in the temporal folder of the system.
        When the context is finished, the folder is automatically deleted."""
        return tempfile.TemporaryDirectory()

    @staticmethod
    def get_tmp_dir():
        """Gets the directory where the temporal
        folder of the system is located."""
        return tempfile.gettempdir()

    @staticmethod
    def get_all_files_in_dir(dir_path):
        """Returns a list with all the file paths in
        the specified directory and subdirectories."""
        files = []
        for dirpath, _, filenames in os.walk(dir_path):
            for filename in filenames:
                files.append(SysUtils.join_paths(dirpath, filename))
        return files

    @staticmethod
    def is_file(file_path):
        """Test whether a path is a regular file."""
        return os.path.isfile(file_path)

    @staticmethod
    def get_file_name(file_path):
        """Returns the filename."""
        return os.path.basename(file_path)

    @staticmethod
    def get_dir_name(file_path):
        """Returns the directory name."""
        return os.path.dirname(file_path)


class StrUtils():
    """Common methods for string management."""

    @staticmethod
    def bytes_to_base64str(value, encoding='utf-8'):
        """Encodes string to base64 and returns another string."""
        return base64.b64encode(value).decode(encoding)

    @staticmethod
    def dict_to_base64str(value, encoding='utf-8'):
        """Encodes a dictionary to base64 and returns a string."""
        return base64.b64encode(json.dumps(value).encode(encoding)).decode(encoding)

    @staticmethod
    def base64_to_str(value, encoding='utf-8'):
        """Decodes from base64 and returns a string."""
        return base64.b64decode(value).decode(encoding)

    @staticmethod
    def utf8_to_base64_string(value):
        """Encode a 'utf-8' string using Base64 and return
        the encoded value as a string."""
        return StrUtils.bytes_to_base64str(value.encode('utf-8'))


class ConfigUtils():
    """Common methods for configuration file management."""

    _LAMBDA_STORAGE_CONFIG_PATH = '/var/task/function_config.yaml'
    _BINARY_STORAGE_CONFIG_ENV = 'FUNCTION_CONFIG'
    _BINARY_OSCAR_STORAGE_CONFIG_PATH = '/oscar/config/function_config.yaml'
    _CUSTOM_VARIABLES = [
        'log_level',
        'execution_mode',
        'extra_payload',
        'udocker_exec',
        'udocker_dir',
        'udocker_bin',
        'udocker_lib'
    ]

    @classmethod
    def read_cfg_var(cls, variable):
        """Returns the value of a config variable or an empty
        string if not found."""
        if SysUtils.is_lambda_environment():
            # Read config file
            with open(cls._LAMBDA_STORAGE_CONFIG_PATH) as file:
                config = yaml.safe_load(file)
        else:
            # Check if config file exsits in '_BINARY_OSCAR_STORAGE_CONFIG_PATH'
            if FileUtils.is_file(cls._BINARY_OSCAR_STORAGE_CONFIG_PATH):
                # Read config file
                with open(cls._BINARY_OSCAR_STORAGE_CONFIG_PATH) as file:
                    config = yaml.safe_load(file)
            else: 
                # Get and decode content of '_BINARY_STORAGE_CONFIG_ENV'
                encoded = SysUtils.get_env_var(cls._BINARY_STORAGE_CONFIG_ENV)
                decoded = StrUtils.base64_to_str(encoded)
                config = yaml.safe_load(decoded)
        # Manage variables that could be defined in environment
        if variable in cls._CUSTOM_VARIABLES:
            value = SysUtils.get_env_var(variable.upper())
            if value != '':
                return value
        return config.get(variable, '') if config else ''



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module in charge of managing the dynamic
data created when a Lambda function is invoked."""

import json
import socket
from faassupervisor.utils import ConfigUtils, FileUtils, StrUtils, SysUtils

def get_function_ip():
    """Returns the IP of the invoked function."""
    return socket.gethostbyname(socket.gethostname())

class LambdaInstance():
    """Stores and manages the Lambda invocation information."""

    PERMANENT_FOLDER = "/var/task"

    def __init__(self, event_info, context):
        self.raw_event = event_info
        self.context = context
        self._set_tmp_folders()
        self._parse_exec_script_and_commands()
        self._set_lambda_env_vars()

    def _set_tmp_folders(self):
        self.input_folder = SysUtils.get_env_var("TMP_INPUT_DIR")
        self.output_folder = SysUtils.get_env_var("TMP_OUTPUT_DIR")

    def _parse_exec_script_and_commands(self):
        # Check for script in function event
        if 'script' in self.raw_event:
            self.script_path = f"{self.input_folder}/script.sh"
            script_content = StrUtils.base64_to_str(self.raw_event['script'])
            FileUtils.create_file_with_content(self.script_path, script_content)
        # Container invoked with arguments
        elif 'cmd_args' in self.raw_event:
            # Add args
            self.cmd_args = json.loads(self.raw_event['cmd_args'])
        # Script to be executed every time (if defined)
        elif ConfigUtils.read_cfg_var('init_script') is not '':
            # Add init script
            self.init_script_path = f"{self.input_folder}/init_script.sh"
            FileUtils.cp_file(ConfigUtils.read_cfg_var('init_script'), self.init_script_path)

    def _set_lambda_env_vars(self):
        SysUtils.set_env_var('AWS_LAMBDA_REQUEST_ID', self.get_request_id())

    def get_memory(self):
        """Returns the amount of memory available to the function in MB."""
        return int(self.context.memory_limit_in_mb)

    def get_request_id(self):
        """Returns the request id of the function invocation."""
        return self.context.aws_request_id

    def get_function_name(self):
        """Returns the name of the function."""
        return self.context.function_name

    def get_log_group_name(self):
        """Returns the name of the Amazon CloudWatch Logs group for the function."""
        return self.context.log_group_name

    def get_log_stream_name(self):
        """Returns the name of the Amazon CloudWatch Logs stream for the function."""
        return self.context.log_stream_name

    def get_function_arn(self):
        """Returns the invoked function ARN."""
        return self.context.invoked_function_arn

    def get_remaining_time_in_seconds(self):
        """Returns the amount of time remaining for the invocation in seconds."""
        remaining_time = int(self.context.get_remaining_time_in_millis() / 1000)
        timeout_threshold = SysUtils.get_env_var('TIMEOUT_THRESHOLD')
        if timeout_threshold is '':
            timeout_threshold = ConfigUtils.read_cfg_var('container')['timeout_threshold']
        return remaining_time - int(timeout_threshold)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the classes and methods
related with the S3 storage provider. """

import boto3
from faassupervisor.logger import get_logger
from faassupervisor.storage.providers import DefaultStorageProvider, \
    get_bucket_name, get_file_key
from faassupervisor.utils import FileUtils, SysUtils


class S3(DefaultStorageProvider):
    """Class that manages downloads and uploads from S3."""

    _TYPE = 'S3'

    def __init__(self, stg_auth):
        super().__init__(stg_auth)
        self.client = self._get_client()

    def _get_client(self):
        """Returns S3 client with default configuration."""
        if self.stg_auth.creds is None:
            return boto3.client('s3')
        region = self.stg_auth.get_credential('region')
        if region == '':
            region = None
        return boto3.client('s3',
                            region_name=region,
                            aws_access_key_id=self.stg_auth.get_credential('access_key'),
                            aws_secret_access_key=self.stg_auth.get_credential('secret_key'))

    def download_file(self, parsed_event, input_dir_path):
        """Downloads the file from the S3 bucket and
        returns the path were the download is placed."""
        file_download_path = SysUtils.join_paths(input_dir_path, parsed_event.file_name)
        get_logger().info('Downloading item from bucket \'%s\' with key \'%s\'',
                          parsed_event.bucket_name,
                          parsed_event.object_key)
        with open(file_download_path, 'wb') as data:
            self.client.download_fileobj(parsed_event.bucket_name,
                                         parsed_event.object_key,
                                         data)
        get_logger().info('Successful download of file \'%s\' from bucket \'%s\' in path \'%s\'',
                          parsed_event.object_key,
                          parsed_event.bucket_name,
                          file_download_path)
        return file_download_path

    def upload_file(self, file_path, file_name, output_path):
        """Uploads the file to the S3 output path."""
        file_key = get_file_key(output_path, file_name)
        bucket_name = get_bucket_name(output_path)
        get_logger().info('Uploading file \'%s\' to bucket \'%s\'', file_key, bucket_name)
        with open(file_path, 'rb') as data:
            self.client.upload_fileobj(data, bucket_name, file_key)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.faas module and classes."""

import unittest
from unittest import mock
import subprocess
from faassupervisor.faas.binary.supervisor import BinarySupervisor
from faassupervisor.faas.aws_lambda.supervisor import LambdaSupervisor, \
                                                      is_batch_execution, \
                                                      _is_lambda_batch_execution
from faassupervisor.exceptions import NoLambdaContextError

# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=protected-access


class BinarySupervisorTest(unittest.TestCase):

    @mock.patch('subprocess.check_output')
    @mock.patch('faassupervisor.utils.FileUtils.create_file_with_content')
    def test_execute_function(self, mock_create, mock_subp):
        supervisor = BinarySupervisor()
        with mock.patch.dict('os.environ', {'SCRIPT':'ZmFrZSBzY3JpcHQh',
                                            'TMP_INPUT_DIR': '/tmp/input'}, clear=True):
            supervisor.execute_function()
            # Check script file creation
            mock_create.assert_called_once_with('/tmp/input/script.sh', 'fake script!')
            # Check process execution
            mock_subp.assert_called_once_with(['/bin/sh', '/tmp/input/script.sh'],
                                              stderr=subprocess.STDOUT)


class LambdaSupervisorTest(unittest.TestCase):

    def _get_context(self):
        mock_context = mock.Mock()
        type(mock_context).aws_request_id = mock.PropertyMock(return_value='123')
        return  mock_context

    def test_is_batch_execution(self):
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'batch'}, clear=True):
            self.assertTrue(is_batch_execution())
        with mock.patch.dict('os.environ', {}, clear=True):
            self.assertFalse(is_batch_execution())

    def test_is_lambda_batch_execution(self):
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'lambda-batch'}, clear=True):
            self.assertTrue(_is_lambda_batch_execution())
        with mock.patch.dict('os.environ', {}, clear=True):
            self.assertFalse(_is_lambda_batch_execution())

    def test_create_lambda_supervisor(self):
        with self.assertRaises(NoLambdaContextError):
            LambdaSupervisor(None, None)
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'lambda-batch',
                                            'TMP_INPUT_DIR': '/tmp/input',
                                            'TMP_OUTPUT_DIR': '/tmp/output'}, clear=True):

            LambdaSupervisor('event', self._get_context())



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module to define classes and methods related
with the API Gateway event.

API Gateway event example:

{'body': '/9j/4AAQSkZJRgAB5Wp//Z',
 'headers': {'Accept': '*/*',
             'Accept-Encoding': 'gzip, deflate',
             'Content-Type': 'application/octet-stream',
             'Host': 'xxxxxx.execute-api.us-east-1.amazonaws.com',
             'User-Agent': 'python-requests/2.21.0',
             'X-Amzn-Trace-Id': 'Root=1-xxx-xxxxx',
             'X-Forwarded-For': '84.111.4.22',
             'X-Forwarded-Port': '443',
             'X-Forwarded-Proto': 'https'},
 'httpMethod': 'POST',
 'isBase64Encoded': False,
 'multiValueHeaders': {'Accept': ['*/*'],
                       'Accept-Encoding': ['gzip, deflate'],
                       'Content-Type': ['application/octet-stream'],
                       'Host': ['xxxxxx.execute-api.us-east-1.amazonaws.com'],
                       'User-Agent': ['python-requests/2.21.0'],
                       'X-Amzn-Trace-Id': ['Root=1-xxx-xxxxx'],
                       'X-Forwarded-For': ['84.111.4.22'],
                       'X-Forwarded-Port': ['443'],
                       'X-Forwarded-Proto': ['https']},
 'multiValueQueryStringParameters': None,
 'path': '/launch',
 'pathParameters': {'proxy': 'launch'},
 'queryStringParameters': None,
 'requestContext': {'accountId': '123456789012',
                    'apiId': 'xxxxxx',
                    'domainName': 'xxxxxx.execute-api.us-east-1.amazonaws.com',
                    'domainPrefix': 'xxxxxx',
                    'extendedRequestId': 'VjR22Hr2oAMFnIw=',
                    'httpMethod': 'POST',
                    'identity': {'accessKey': None,
                                 'accountId': None,
                                 'caller': None,
                                 'cognitoAuthenticationProvider': None,
                                 'cognitoAuthenticationType': None,
                                 'cognitoIdentityId': None,
                                 'cognitoIdentityPoolId': None,
                                 'sourceIp': '84.111.4.22',
                                 'user': None,
                                 'userAgent': 'python-requests/2.21.0',
                                 'userArn': None},
                    'path': '/scar/launch',
                    'protocol': 'HTTP/1.1',
                    'requestId': '385ed319-375f-11e9-aa95-df027d8eef05',
                    'requestTime': '23/Feb/2019:11:36:11 +0000',
                    'requestTimeEpoch': 1550921771868,
                    'resourceId': 'asdfg',
                    'resourcePath': '/{proxy+}',
                    'stage': 'scar'},
 'resource': '/{proxy+}',
 'stageVariables': None}
"""

import base64
from faassupervisor.utils import FileUtils, SysUtils
from faassupervisor.events.unknown import UnknownEvent

_JSON_TYPE = 'application/json'


class ApiGatewayEvent(UnknownEvent):
    """ Parse the API Gateway event and saves the body
    (if exists) and the request parameters (if exists). """

    _TYPE = 'APIGATEWAY'
    _FILE_NAME = 'api_event_file'

    def _set_event_params(self):
        """If has a JSON body, it can contain
        storage provider information so we save
        it for further parsing.

        Also check for request parameters."""
        self.body = self.event.get('body')
        if not self.body:
            self.body = ''
        if self._is_request_with_parameters():
            self._save_request_parameters()

    def has_json_body(self):
        """Returns true if the type of the request is JSON"""
        if ('Content-Type' in self.event['headers'] and
                self.event['headers']['Content-Type']):
            return self.event['headers']['Content-Type'].strip() == _JSON_TYPE
        return False

    def _is_request_with_parameters(self):
        return "queryStringParameters" in self.event \
                and self.event["queryStringParameters"]

    def _save_request_parameters(self):
        # Add passed HTTP parameters to container variables
        for key, value in self.event["queryStringParameters"].items():
            SysUtils.set_env_var(f"CONT_VAR_{key}", value)

    def save_event(self, input_dir_path):
        file_path = SysUtils.join_paths(input_dir_path, self._FILE_NAME)
        if self.has_json_body():
            FileUtils.create_file_with_content(file_path, self.body)
        else:
            FileUtils.create_file_with_content(file_path,
                                               base64.b64decode(self.body),
                                               mode='wb')
        return file_path



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to parse, store and manage storage information."""
from faassupervisor.utils import ConfigUtils, FileUtils
from faassupervisor.exceptions import StorageAuthError, \
    InvalidStorageProviderError, exception
from faassupervisor.storage.providers.local import Local
from faassupervisor.storage.providers.minio import Minio
from faassupervisor.storage.providers.onedata import Onedata
from faassupervisor.storage.providers.s3 import S3
from faassupervisor.logger import get_logger


def create_provider(storage_auth):
    """Returns the storage provider needed
    based on the authentication type defined.

    If not storage auth provided, use local storage."""
    if not storage_auth:
        provider = Local(storage_auth)
    elif storage_auth.type == 'MINIO':
        provider = Minio(storage_auth)
    elif storage_auth.type == 'ONEDATA':
        provider = Onedata(storage_auth)
    elif storage_auth.type == 'S3':
        provider = S3(storage_auth)
    else:
        raise InvalidStorageProviderError(storage_type=storage_auth.type)
    return provider


class AuthData():
    """Stores provider authentication values."""

    def __init__(self, storage_type, credentials):
        self.type = storage_type
        self.creds = credentials

    def get_credential(self, key):
        """Return authentication credentials previously stored."""
        return self.creds.get(key, '')


class StorageConfig():
    """Parses providers authentication variables and defined outputs."""

    def __init__(self):
        # Create s3_auth with empty credentials
        self.s3_auth = AuthData('S3', None)
        self.minio_auth = None
        self.onedata_auth = None
        self.output = []
        self._parse_config()

    @exception()
    def _parse_config(self):
        output = ConfigUtils.read_cfg_var('output')
        # Output list
        if output != '':
            self.output = output
        else:
            get_logger().warning('There is no output defined for this function.')
        storage_providers = ConfigUtils.read_cfg_var('storage_providers')
        if (storage_providers and
                storage_providers is not ''):
            # s3 storage provider auth
            if ('s3' in storage_providers
                    and storage_providers['s3']):
                self._validate_s3_creds(storage_providers['s3'])
            # minio storage provider auth
            if ('minio' in storage_providers
                    and storage_providers['minio']):
                self._validate_minio_creds(storage_providers['minio'])
            # onedata storage provider auth
            if ('onedata' in storage_providers
                    and storage_providers['onedata']):
                self._validate_onedata_creds(storage_providers['onedata'])
        else:
            get_logger().warning('There is no storage provider defined for this function.')

    def _validate_minio_creds(self, minio_creds):
        if (isinstance(minio_creds, dict)
                and 'access_key' in minio_creds
                and minio_creds['access_key'] is not None
                and minio_creds['access_key'] != ''
                and 'secret_key' in minio_creds
                and minio_creds['secret_key'] is not None
                and minio_creds['secret_key'] != ''):
            self.minio_auth = AuthData('MINIO', minio_creds)
        else:
            raise StorageAuthError(auth_type='MINIO')

    def _validate_s3_creds(self, s3_creds):
        if (isinstance(s3_creds, dict)
                and 'access_key' in s3_creds
                and s3_creds['access_key'] is not None
                and s3_creds['access_key'] != ''
                and 'secret_key' in s3_creds
                and s3_creds['secret_key'] is not None
                and s3_creds['secret_key'] != ''):
            self.s3_auth = AuthData('S3', s3_creds)
        else:
            raise StorageAuthError(auth_type='S3')

    def _validate_onedata_creds(self, onedata_creds):
        if (isinstance(onedata_creds, dict)
                and 'oneprovider_host' in onedata_creds
                and onedata_creds['oneprovider_host'] is not None
                and onedata_creds['oneprovider_host'] != ''
                and 'token' in onedata_creds
                and onedata_creds['token'] is not None
                and onedata_creds['token'] != ''
                and 'space' in onedata_creds
                and onedata_creds['space'] is not None
                and onedata_creds['space'] != ''):
            self.onedata_auth = AuthData('ONEDATA', onedata_creds)
        else:
            raise StorageAuthError(auth_type='ONEDATA')

    def get_auth_data_by_stg_type(self, storage_type):
        """Returns the authentication credentials by its type."""
        if storage_type == 'S3':
            return self.s3_auth
        elif storage_type == 'MINIO':
            return self.minio_auth
        elif storage_type == 'ONEDATA':
            return self.onedata_auth
        return None

    def download_input(self, parsed_event, input_dir_path):
        """Receives the event where the file information is and
        the tmp_dir_path where to store the downloaded file.

        Returns the file path where the file is downloaded."""
        event_type = parsed_event.get_type()
        auth_data = self.get_auth_data_by_stg_type(event_type)
        stg_provider = create_provider(auth_data)
        get_logger().info('Found \'%s\' input provider', stg_provider.get_type())
        return stg_provider.download_file(parsed_event, input_dir_path)

    def upload_output(self, output_dir_path):
        """Receives the tmp_dir_path where the files to upload are stored and
        uploads files whose name matches the prefixes and suffixes specified
        in 'output'."""
        get_logger().info('Searching for files to upload in folder \'%s\'', output_dir_path)
        output_files = FileUtils.get_all_files_in_dir(output_dir_path)
        stg_providers = {}
        # Filter files by prefix and suffix
        for output in self.output:
            get_logger().info('Checking files for uploading to \'%s\' on path: \'%s\'',
                              output['storage_provider'].upper(),
                              output['path'])
            for file_path in output_files:
                file_name = file_path.replace(f'{output_dir_path}/', '')
                prefix_ok = False
                suffix_ok = False
                # Check prefixes
                if ('prefix' not in output
                        or len(output['prefix']) == 0):
                    prefix_ok = True
                else:
                    for pref in output['prefix']:
                        if file_name.startswith(pref):
                            prefix_ok = True
                            break
                if prefix_ok:
                    # Check suffixes
                    if ('suffix' not in output
                            or len(output['suffix']) == 0):
                        suffix_ok = True
                    else:
                        for suff in output['suffix']:
                            if file_name.endswith(suff):
                                suffix_ok = True
                                break
                    # Only upload file if name matches the prefixes and suffixes
                    if suffix_ok:
                        out_type = output['storage_provider'].upper()
                        if out_type not in stg_providers:
                            auth_data = self.get_auth_data_by_stg_type(out_type)
                            stg_providers[out_type] = create_provider(auth_data)
                        stg_providers[out_type].upload_file(file_path, file_name, output['path'])



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module used to define a generic unknown event."""

import json
import base64
import uuid
from faassupervisor.utils import SysUtils, FileUtils


class UnknownEvent():
    """Class to manage unknown events."""

    _TYPE = 'UNKNOWN'

    def __init__(self, event):
        self._file_name = 'event-file-{}'.format(str(uuid.uuid4()))
        self.event = event
        if isinstance(event, dict):
            records = event.get('Records')
            if records:
                self.event_records = records[0]
        self._set_event_params()

    def _set_event_params(self):
        """ Generic method to be implemented by all the event parsers. """

    def get_type(self):
        """Returns the event type.
        Default event is UNKNOWN, but it can also
        be APIGATEWAY, MINIO, ONEDATA, and S3.

        Each class inheriting from UnkownEvent
        must override the _TYPE."""
        return self._TYPE

    def save_event(self, input_dir_path):
        """Stores the unknown event and returns
        the file path where the file is stored."""
        file_path = SysUtils.join_paths(input_dir_path, self._file_name)
        try:
            json.loads(self.event)
        except ValueError:
            FileUtils.create_file_with_content(file_path,
                                               base64.b64decode(self.event),
                                               mode='wb')
        except TypeError:
            FileUtils.create_file_with_content(file_path, self.event)
        else:
            FileUtils.create_file_with_content(file_path, self.event)

        return file_path



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_namespace_packages

# Load readme
with open('README.md', mode='r', encoding='utf-8') as f:
    readme = f.read()

# Load version
with open('faassupervisor/version.py', mode='r', encoding='utf-8') as f:
    exec(f.read())

setup(name='faas-supervisor',
      version=__version__,
      description='Input/Output data manager for FaaS',
      long_description=readme,
      long_description_content_type='text/markdown',
      url='https://github.com/grycap/faas-supervisor',
      author='GRyCAP - Universitat Politecnica de Valencia',
      author_email='alpegon3@upv.es',
      license='Apache 2.0',
      packages=find_namespace_packages(),
      install_requires=[
          'requests',
          'boto3',
          'setuptools >= 40.8.0',
          'pyyaml'
      ],
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: Apache Software License'
      ],
      zip_safe=False)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Minio event example:

{"Key": "images/nature-wallpaper-229.jpg",
 "Records": [{"s3": {"object": {"key": "nature-wallpaper-229.jpg",
                                "userMetadata": { "content-type": "image/jpeg"},
                                "eTag": "dd20b7e4b74467ff16ce2d901c054419",
                                "contentType": "image/jpeg",
                                "sequencer": "153C9A7A7A3FB6AE",
                                "versionId": "1",
                                "size": 1019645},
                     "s3SchemaVersion": "1.0",
                     "bucket": {"ownerIdentity": {"principalId": "minio"},
                                "name": "images",
                                "arn": "arn:aws:s3:::images"},
                     "configurationId": "Config"},
              "requestParameters": {"sourceIPAddress": "10.244.0.0:34852"},
              "responseElements": {"x-amz-request-id": "153C9A7A7A3FB6AE",
                                   "x-minio-origin-endpoint": "http://10.244.1.3:9000"},
              "source": {"userAgent": "",
                         "host": "",
                         "port": ""},
              "eventVersion": "2.0",
              "eventName": "s3:ObjectCreated:Put",
              "awsRegion": "",
              "eventTime": "2018-06-29T10:23:44Z",
              "eventSource": "minio:s3",
              "userIdentity": {"principalId": "minio"}}],
 "EventName": "s3:ObjectCreated:Put"}
"""

from urllib.parse import unquote_plus
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.utils import FileUtils


class MinioEvent(UnknownEvent):
    """Class to parse the Minio event."""

    # pylint: disable=too-few-public-methods

    _TYPE = 'MINIO'

    def _set_event_params(self):
        self.bucket_arn = self.event_records['s3']['bucket']['arn']
        self.bucket_name = self.event_records['s3']['bucket']['name']
        self.object_key = unquote_plus(self.event_records['s3']['object']['key'])
        self.file_name = FileUtils.get_file_name(self.object_key)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.storage module and classes."""

import unittest
from unittest import mock
from unittest.mock import call
from collections import namedtuple
from faassupervisor.storage.providers import get_bucket_name, get_file_key
from faassupervisor.storage.config import StorageConfig, AuthData, create_provider
from faassupervisor.storage.providers.local import Local
from faassupervisor.storage.providers.minio import Minio
from faassupervisor.storage.providers.onedata import Onedata
from faassupervisor.storage.providers.s3 import S3
from faassupervisor.events.minio import MinioEvent
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.events.s3 import S3Event
from faassupervisor.events.onedata import OnedataEvent
from faassupervisor.utils import StrUtils

# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=protected-access

CONFIG_FILE_OK = """
name: test-func
output:
- storage_provider: s3
  path: bucket/folder
- storage_provider: minio
  path: bucket
  suffix: ['txt', 'jpg']
  prefix: ['result-']
storage_providers:
  minio:
    access_key: test_minio_access
    secret_key: test_minio_secret
  onedata:
    oneprovider_host: test_oneprovider.host
    token: test_onedata_token
    space: test_onedata_space
"""

CONFIG_FILE_NO_OUTPUT = """
name: test-func
storage_providers:
  minio:
    access_key: test_minio_access
    secret_key: test_minio_secret
"""

CONFIG_FILE_NO_STORAGE_PROVIDER = """
name: test-func
output:
- storage_provider: s3
  path: bucket/folder
"""

CONFIG_FILE_INVALID_MINIO = """
name: test-func
storage_providers:
  minio:
    secret_key: test_minio_secret
"""

CONFIG_FILE_INVALID_ONEDATA = """
name: test-func
storage_providers:
  onedata:
    token: test_onedata_token
    space: test_onedata_space
"""

CONFIG_FILE_INVALID_S3 = """
name: test-func
storage_providers:
  s3:
    user: test
"""

class AuthDataTest(unittest.TestCase):

    def test_create_auth_data(self):
        auth = AuthData('MINIO', {'access_key': 'test'})
        self.assertEqual(auth.type, 'MINIO')
        self.assertEqual(auth.creds, {'access_key': 'test'})

    def test_get_auth_data_credential(self):
        auth = AuthData('MINIO', {'access_key': 'test'})
        self.assertEqual(auth.get_credential('access_key'), 'test')
        self.assertEqual(auth.get_credential('k'), '')


class StorageConfigTest(unittest.TestCase):

    def test_parse_config_valid(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            config = StorageConfig()
            expected_output = [
                {
                    'storage_provider': 's3',
                    'path': 'bucket/folder'
                }, {
                    'storage_provider': 'minio',
                    'path': 'bucket',
                    'suffix': ['txt', 'jpg'],
                    'prefix': ['result-']
                }
            ]
            self.assertEqual(config.output, expected_output)
            self.assertEqual(config.minio_auth.type, 'MINIO')
            self.assertEqual(config.minio_auth.get_credential('access_key'), 'test_minio_access')
            self.assertEqual(config.minio_auth.get_credential('secret_key'), 'test_minio_secret')
            self.assertEqual(config.onedata_auth.type, 'ONEDATA')
            self.assertEqual(config.onedata_auth.get_credential('oneprovider_host'), 'test_oneprovider.host')
            self.assertEqual(config.onedata_auth.get_credential('token'), 'test_onedata_token')
            self.assertEqual(config.onedata_auth.get_credential('space'), 'test_onedata_space')
            self.assertEqual(config.s3_auth.type, 'S3')

    def test_parse_config_no_output(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_NO_OUTPUT)},
                             clear=True):
            StorageConfig()
            self.assertLogs('There is no output defined for this function.',
                             level='WARNING')

    def test_parse_config_no_storage_provider(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_NO_STORAGE_PROVIDER)},
                             clear=True):
            StorageConfig()
            self.assertLogs('There is no storage provider defined for this function.',
                             level='WARNING')

    def test_parse_config_invalid_minio(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_INVALID_MINIO)},
                             clear=True):
            with self.assertRaises(SystemExit):
                StorageConfig()
                self.assertLogs('The storage authentication of \'MINIO\' is not well-defined.',
                                level='ERROR')

    def test_parse_config_invalid_s3(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_INVALID_S3)},
                             clear=True):
            with self.assertRaises(SystemExit):
                StorageConfig()
                self.assertLogs('The storage authentication of \'S3\' is not well-defined.',
                                level='ERROR')

    def test_parse_config_invalid_onedata(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_INVALID_ONEDATA)},
                             clear=True):
            with self.assertRaises(SystemExit):
                StorageConfig()
                self.assertLogs('The storage authentication of \'ONEDATA\' is not well-defined.',
                                level='ERROR')

    def test_get_minio_auth(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            minio_auth = StorageConfig().get_auth_data_by_stg_type('MINIO')
            self.assertEqual(minio_auth.type, 'MINIO')
            self.assertEqual(minio_auth.get_credential('access_key'),
                             'test_minio_access')
            self.assertEqual(minio_auth.get_credential('secret_key'),
                             'test_minio_secret')

    def test_get_s3_auth(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            s3_auth = StorageConfig().get_auth_data_by_stg_type('S3')
            self.assertEqual(s3_auth.type, 'S3')

    def test_get_onedata_auth(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            onedata_auth = StorageConfig().get_auth_data_by_stg_type('ONEDATA')
            self.assertEqual(onedata_auth.type, 'ONEDATA')
            self.assertEqual(onedata_auth.get_credential('oneprovider_host'),
                             'test_oneprovider.host')
            self.assertEqual(onedata_auth.get_credential('token'),
                             'test_onedata_token')
            self.assertEqual(onedata_auth.get_credential('space'),
                             'test_onedata_space')

    def test_get_invalid_auth(self):
        invalid_auth = StorageConfig().get_auth_data_by_stg_type('INVALID_TYPE')
        self.assertIsNone(invalid_auth)

    @mock.patch('faassupervisor.storage.providers.local.Local.download_file')
    def test_download_input(self, mock_download_file):
        event = UnknownEvent('')
        StorageConfig().download_input(event, '/tmp/test')
        mock_download_file.assert_called_once_with(event, '/tmp/test')

    @mock.patch('faassupervisor.utils.FileUtils.get_all_files_in_dir')
    @mock.patch('faassupervisor.storage.providers.s3.S3.upload_file')
    @mock.patch('faassupervisor.storage.providers.minio.Minio.upload_file')
    def test_upload_output(self, mock_minio, mock_s3, mock_get_files):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            files = [
                '/tmp/test/file1.txt',
                '/tmp/test/file1.jpg',
                '/tmp/test/result-file.txt',
                '/tmp/test/result-file.out',
                '/tmp/test/file2.txt',
                '/tmp/test/file2.out'
            ]
            mock_get_files.return_value = files
            StorageConfig().upload_output('/tmp/test')
            self.assertEqual(mock_minio.call_count, 1)
            self.assertEqual(mock_minio.call_args,
                             call('/tmp/test/result-file.txt',
                                  'result-file.txt',
                                  'bucket'))
            self.assertEqual(mock_s3.call_count, 6)
            for i, f in enumerate(files):
                self.assertEqual(mock_s3.call_args_list[i],
                                 call(f, f.split('/')[3], 'bucket/folder'))


class ProvidersModuleTest(unittest.TestCase):

    def test_get_bucket_name(self):
        self.assertEqual(get_bucket_name('bucket/folder1/folder2'), 'bucket')
        self.assertEqual(get_bucket_name('bucket'), 'bucket')

    def test_get_file_key(self):
        self.assertEqual(get_file_key('bucket/folder1/folder2', 'file'),
                         'folder1/folder2/file')
        self.assertEqual(get_file_key('bucket', 'file'), 'file')


class LocalProviderTest(unittest.TestCase):

    def test_create_local_provider(self):
        provider = create_provider(None)
        self.assertEqual(provider.get_type(), 'LOCAL')

    def test_download_file(self):
        provider = Local(None)
        parsed_event = mock.Mock(spec=UnknownEvent)
        provider.download_file(parsed_event, '/tmp/local')
        parsed_event.save_event.assert_called_once_with('/tmp/local')


class MinioProviderTest(unittest.TestCase):

    MINIO_CREDS = {
            'access_key': 'test_minio_access',
            'secret_key': 'test_minio_secret'
    }

    def test_create_minio_provider(self):
        minio_auth = AuthData('MINIO', self.MINIO_CREDS)
        provider = create_provider(minio_auth)
        self.assertEqual(provider.get_type(), 'MINIO')
        self.assertEqual(provider.stg_auth.creds, self.MINIO_CREDS)

    @mock.patch('boto3.client')
    def test_get_client_default_endpoint(self, mock_boto):
        Minio(AuthData('MINIO', self.MINIO_CREDS))
        mock_boto.assert_called_once_with('s3',
                                           endpoint_url='http://minio-service.minio:9000',
                                           region_name=None,
                                           verify=True,
                                           aws_access_key_id='test_minio_access',
                                           aws_secret_access_key='test_minio_secret')

    @mock.patch('boto3.client')
    def test_get_client_custom_endpoint(self, mock_boto):
        Minio(AuthData('MINIO', {**self.MINIO_CREDS,
                                 'endpoint': 'https://test.endpoint'}))
        mock_boto.assert_called_once_with('s3',
                                           endpoint_url='https://test.endpoint',
                                           region_name=None,
                                           verify=True,
                                           aws_access_key_id='test_minio_access',
                                           aws_secret_access_key='test_minio_secret')

    @mock.patch('boto3.client')
    def test_download_file(self, mock_boto):
        minio_provider = Minio(AuthData('MINIO', self.MINIO_CREDS))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            # Create mock event and event properties
            event = mock.Mock(spec=MinioEvent)
            type(event).bucket_name = mock.PropertyMock(return_value='minio_bucket')
            type(event).file_name = mock.PropertyMock(return_value='minio_file')
            type(event).object_key = mock.PropertyMock(return_value='minio_file')
            file_path = minio_provider.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/minio_file')
            # Check file writing call
            mopen.assert_called_once_with('/tmp/input/minio_file', 'wb')
            # Check boto client download call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().download_fileobj('minio_bucket',
                                                     'minio_file',
                                                     mopen.return_value))

    @mock.patch('boto3.client')
    def test_upload_file(self, mock_boto):
        minio_provider = Minio(AuthData('MINIO', self.MINIO_CREDS))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            minio_provider.upload_file('/tmp/output/processed.jpg', 'processed.jpg', 'minio_bucket')
            # Check file reading call
            mopen.assert_called_once_with('/tmp/output/processed.jpg', 'rb')
            # Check boto client upload call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().upload_fileobj(mopen.return_value,
                                                   'minio_bucket',
                                                   'processed.jpg'))


class OnedataProviderTest(unittest.TestCase):

    ONEDATA_CREDS = {
        'oneprovider_host': 'test_oneprovider.host',
        'token': 'test_onedata_token',
        'space': 'test_onedata_space'
    }
    def test_create_onedata_provider(self):
        onedata_auth = AuthData('ONEDATA', self.ONEDATA_CREDS)
        provider = create_provider(onedata_auth)
        self.assertEqual(provider.get_type(), 'ONEDATA')
        self.assertEqual(provider.stg_auth.creds, self.ONEDATA_CREDS)

    @mock.patch('requests.get')
    def test_download_file(self, mock_requests):
        onedata_provider = Onedata(AuthData('ONEDATA', self.ONEDATA_CREDS))
        # Mock requests.get response
        response = namedtuple('response', ['content', 'status_code'])
        mock_requests.return_value = response('test response', 200)
        # Create mock event
        event = mock.Mock(spec=OnedataEvent)
        type(event).file_name = mock.PropertyMock(return_value='onedata_file')
        type(event).object_key = mock.PropertyMock(return_value='/onedata_file_key')
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            file_path = onedata_provider.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/onedata_file')
            # Check request to onedata endpoint
            mock_requests.assert_called_once_with('https://test_oneprovider.host/cdmi/onedata_file_key',
                                                  headers={'X-Auth-Token': 'test_onedata_token'})
            # Check file writing
            mopen.assert_called_once_with('/tmp/input/onedata_file', 'wb')

    @mock.patch('requests.get')
    @mock.patch('requests.put')
    def test_upload_file(self, mock_put, mock_get):
        onedata_provider = Onedata(AuthData('ONEDATA', self.ONEDATA_CREDS))
        response = namedtuple('response', ['status_code'])
        # Mock requests.get response (check if folder exists)
        mock_get.return_value = response(200)
        # Mock requests.put response
        mock_put.return_value = response(202)
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            onedata_provider.upload_file('/tmp/output/onedata_file', 'onedata_file', 'onedata_path')
            # Check request to onedata endpoint
            mock_get.assert_called_once_with(
                'https://test_oneprovider.host/cdmi/test_onedata_space/onedata_path/',
                headers={**onedata_provider._CDMI_VERSION_HEADER,
                         'X-Auth-Token': 'test_onedata_token'})
            mock_put.assert_called_once_with(
                'https://test_oneprovider.host/cdmi/test_onedata_space/onedata_path/onedata_file',
                data=mopen.return_value,
                headers={'X-Auth-Token': 'test_onedata_token'})
            # Check file writing
            mopen.assert_called_once_with('/tmp/output/onedata_file', 'rb')


class S3ProviderTest(unittest.TestCase):

    S3_CREDS = {
        'access_key': 'test_s3_access',
        'secret_key': 'test_s3_secret'
    }

    def test_create_s3_provider(self):
        minio_auth = AuthData('S3', None)
        provider = create_provider(minio_auth)
        self.assertEqual(provider.get_type(), 'S3')
        self.assertIsNone(provider.stg_auth.creds)

    @mock.patch('boto3.client')
    def test_get_client_without_creds(self, mock_boto):
        S3(AuthData('S3', None))
        mock_boto.assert_called_once_with('s3')

    @mock.patch('boto3.client')
    def test_get_client_with_creds(self, mock_boto):
        S3(AuthData('S3', self.S3_CREDS))
        mock_boto.assert_called_once_with('s3',
                                           region_name=None,
                                           aws_access_key_id='test_s3_access',
                                           aws_secret_access_key='test_s3_secret')

    @mock.patch('boto3.client')
    def test_download_file(self, mock_boto):
        s3_provider = S3(AuthData('S3', None))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            # Create mock event and event properties
            event = mock.Mock(spec=S3Event)
            type(event).bucket_name = mock.PropertyMock(return_value='s3_bucket')
            type(event).file_name = mock.PropertyMock(return_value='s3_file')
            type(event).object_key = mock.PropertyMock(return_value='s3_folder/s3_file')
            file_path = s3_provider.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/s3_file')
            # Check file writing call
            mopen.assert_called_once_with('/tmp/input/s3_file', 'wb')
            # Check boto client download call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().download_fileobj('s3_bucket',
                                                     's3_folder/s3_file',
                                                     mopen.return_value))

    @mock.patch('boto3.client')
    def test_upload_file(self, mock_boto):
        s3_provider = S3(AuthData('S3', None))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            s3_provider.upload_file('/tmp/output/processed.jpg', 'processed.jpg', 's3_bucket/s3_folder')
            # Check file reading call
            mopen.assert_called_once_with('/tmp/output/processed.jpg', 'rb')
            # Check boto client upload call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().upload_fileobj(mopen.return_value,
                                                   's3_bucket',
                                                   's3_folder/processed.jpg'))



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.utils module."""

import sys
import io
import os
import unittest
from unittest import mock
from faassupervisor.utils import SysUtils, StrUtils, FileUtils, ConfigUtils

# pylint: disable=missing-docstring
# pylint: disable=no-self-use

CONFIG_FILE = """
name: test-func
"""

class SysUtilsTest(unittest.TestCase):

    def test_get_stdin(self):
        sys.stdin = io.StringIO('It works!')
        self.assertEqual(SysUtils.get_stdin(), 'It works!')

    def test_join_paths(self):
        paths = ['It', 'works', '!']
        # Works in Linux systems
        self.assertEqual(SysUtils.join_paths(*paths), "/".join(paths))

    def test_is_var_in_env(self):
        with mock.patch.dict('os.environ', {"K1":"V1"}, clear=True):
            self.assertTrue(SysUtils.is_var_in_env('K1'))

    def test_set_env_var(self):
        with mock.patch.dict('os.environ', {}, clear=True):
            SysUtils.set_env_var('TEST_VAR', 'TEST_VAL')
            self.assertTrue('TEST_VAR' in os.environ)
            self.assertEqual(os.environ['TEST_VAR'], 'TEST_VAL')

    def test_get_all_env_vars(self):
        with mock.patch.dict('os.environ', {"K1":"V1", "K2":"V2"}, clear=True):
            self.assertEqual(SysUtils.get_all_env_vars(), {"K1":"V1", "K2":"V2"})

    def test_get_env_var(self):
        with mock.patch.dict('os.environ', {"K1":"V1", "K2":"V2"}, clear=True):
            self.assertEqual(SysUtils.get_env_var("K2"), "V2")
            self.assertEqual(SysUtils.get_env_var("K3"), "")

#     def test_get_cont_env_vars(self):
#         with mock.patch.dict('os.environ',
#                              {"K1":"V1", "CONT_VAR_C1":"VC1", "CONT_VAR_C2":"VC2"},
#                              clear=True):
#             # Variables without the prefix
#             self.assertEqual(SysUtils.get_cont_env_vars(), {"C1":"VC1", "C2":"VC2"})
#  
#         with mock.patch.dict('os.environ', {"K1":"V1"}, clear=True):
#             self.assertEqual(SysUtils.get_cont_env_vars(), {})

    @mock.patch('subprocess.call')
    def test_execute_cmd(self, mock_call):
        SysUtils.execute_cmd(["ls", "-la"])
        mock_call.assert_called_once_with(["ls", "-la"])

    @mock.patch('subprocess.check_output')
    def test_execute_command_and_return_output(self, mock_check_output):
        mock_check_output.return_value = b'testing\nreturn\ndecode.'
        out = SysUtils.execute_cmd_and_return_output(["ls", "-la"])
        mock_check_output.assert_called_once_with(["ls", "-la"])
        self.assertTrue(out, "testing\nreturn\ndecode")


class FileUtilsTest(unittest.TestCase):

    @mock.patch('os.stat')
    @mock.patch('os.chmod')
    def test_set_file_execution_rights(self, mock_chmod, mock_stat):
        mock_stat().st_mode = 0o0000
        FileUtils.set_file_execution_rights('/tmp/invented_file')
        # 73 is = 0o000 | 0o0111 in decimal
        mock_chmod.assert_called_once_with('/tmp/invented_file', 73)

    @mock.patch('shutil.copyfile')
    def test_cp_file(self, mock_cp):
        FileUtils.cp_file('/tmp/src_file', '/tmp/dst_file')
        mock_cp.assert_called_once_with('/tmp/src_file', '/tmp/dst_file')

    @mock.patch('os.makedirs')
    def test_create_folder(self, mock_dir):
        FileUtils.create_folder('/tmp/folder')
        mock_dir.assert_called_once_with('/tmp/folder', exist_ok=True)

    def test_create_file_with_content(self):
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            FileUtils.create_file_with_content('/tmp/file', 'fifayfofum')
            mopen.assert_called_once_with('/tmp/file', 'w')
            mopen().write.assert_called_once_with('fifayfofum')

    def test_create_file_with_json_content(self):
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            FileUtils.create_file_with_content('/tmp/file', {"k1":"v1", "k2":"v2"})
            mopen.assert_called_once_with('/tmp/file', 'w')
            mopen().write.assert_called_once_with('{"k1": "v1", "k2": "v2"}')

    def test_read_file(self):
        mopen = mock.mock_open(read_data='fifayfofum')
        with mock.patch('builtins.open', mopen, create=True):
            content = FileUtils.read_file('/tmp/file')
            mopen.assert_called_once_with('/tmp/file', mode='r', encoding='utf-8')
            self.assertEqual(content, 'fifayfofum')

    def test_read_binary_file(self):
        mopen = mock.mock_open(read_data='fifayfofum')
        with mock.patch('builtins.open', mopen, create=True):
            content = FileUtils.read_file('/tmp/file', file_mode='rb')
            mopen.assert_called_once_with('/tmp/file', mode='rb', encoding=None)
            self.assertEqual(content, 'fifayfofum')

    @mock.patch('tempfile.TemporaryDirectory')
    def test_create_tmp_dir(self, mock_tmp):
        FileUtils.create_tmp_dir()
        mock_tmp.assert_called_once()

    @mock.patch('tempfile.gettempdir')
    def test_get_tmp_dir(self, mock_tmp):
        FileUtils.get_tmp_dir()
        mock_tmp.assert_called_once()

    @mock.patch('os.walk')
    def test_get_all_files_in_dir(self, mock_os):
        mock_os.return_value = [('/tmp', ['t1'], ['f1', 'f2']),
                                ('/tmp/t1', [], ['f3'])]
        files = FileUtils.get_all_files_in_dir('/tmp')
        mock_os.assert_called_once_with('/tmp')
        self.assertEqual(files, ['/tmp/f1', '/tmp/f2', '/tmp/t1/f3'])

    @mock.patch('os.path.isfile')
    def test_is_file(self, mock_os):
        FileUtils.is_file('/tmp/invented_file')
        mock_os.assert_called_with('/tmp/invented_file')

    @mock.patch('os.path.basename')
    def test_get_file_name(self, mock_os):
        FileUtils.get_file_name('/tmp/invented_file.jpg')
        mock_os.assert_called_with('/tmp/invented_file.jpg')


class StrUtilsTest(unittest.TestCase):

    def test_bytes_to_base64str(self):
        self.assertEqual(StrUtils.bytes_to_base64str(b'testing\nencode.'), "dGVzdGluZwplbmNvZGUu")

    def test_dict_to_base64str(self):
        self.assertEqual(StrUtils.dict_to_base64str({"k1":"v1", "k2":"v2"}),
                         "eyJrMSI6ICJ2MSIsICJrMiI6ICJ2MiJ9")

    def test_base64_to_str(self):
        self.assertEqual(StrUtils.base64_to_str("dGVzdGluZwplbmNvZGUu"), "testing\nencode.")

    def test_utf8_to_base64_string(self):
        self.assertEqual(StrUtils.utf8_to_base64_string("testing\nencode."), "dGVzdGluZwplbmNvZGUu")


class ConfigUtilsTest(unittest.TestCase):

    def test_read_cfg_var_environment(self):
        with mock.patch.dict('os.environ', {'LOG_LEVEL': 'TEST'}, clear=True):
            self.assertEqual(ConfigUtils.read_cfg_var('log_level'), 'TEST')

    def test_read_cfg_var_config_file(self):
        with mock.patch.dict('os.environ',
                             {'AWS_EXECUTION_ENV': 'AWS_Lambda_'},
                             clear=True):
            mopen = mock.mock_open(read_data=CONFIG_FILE)
            with mock.patch('builtins.open', mopen, create=True):
                var = ConfigUtils.read_cfg_var('name')
                mopen.assert_called_once_with('/var/task/function_config.yaml')
                self.assertEqual(var, 'test-func')

    def test_read_cfg_var_config_encoded(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE)},
                             clear=True):
            self.assertEqual(ConfigUtils.read_cfg_var('name'), 'test-func')



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.events module and classes."""

import os
import unittest
from unittest import mock
import faassupervisor.events as events
from faassupervisor.events.s3 import S3Event
from faassupervisor.events.minio import MinioEvent
from faassupervisor.events.onedata import OnedataEvent
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.events.apigateway import ApiGatewayEvent
from faassupervisor.exceptions import UnknowStorageEventWarning

# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=protected-access

MINIO_EVENT = {"Key": "images/nature-wallpaper-229.jpg",
               "Records": [{"s3": {"object": {"key": "nature-wallpaper-229.jpg"},
                                   "bucket": {"name": "images",
                                              "arn": "arn:aws:s3:::images"}},
                            "eventSource": "minio:s3"}]}

ONEDATA_EVENT = {"Key": "/my-onedata-space/files/file.txt",
                 "Records": [{"objectKey": "file.txt",
                              "eventSource": "OneTrigger"}]}

S3_EVENT = {'Records': [{'awsRegion': 'us-east-1',
                         'eventSource': 'aws:s3',
                         's3': {'bucket': {'arn': 'arn:aws:s3:::darknet-bucket',
                                           'name': 'darknet-bucket'},
                                'object': {'key': 'darknet-s3/input/dog.jpg'}}}]}

UNKNOWN_EVENT = {'Records': [{'eventSource': 'narnia'}]}

APIGTW_EVENT_WO_JSON = {'body': 'aXQgd29ya3Mh',
                        'headers': {'Content-Type': 'application/octet-stream'},
                        'httpMethod': 'POST',
                        'isBase64Encoded': False,
                        'queryStringParameters': None}

APIGTW_EVENT_W_JSON = {'body': S3_EVENT,
                       'headers': {'Content-Type': 'application/json'},
                       'httpMethod': 'POST',
                       'isBase64Encoded': False,
                       'queryStringParameters': {'q1':'v1', 'q2':'v2'}}


class EventModuleTest(unittest.TestCase):

    def test_is_api_gateway_event_true(self):
        self.assertTrue(events._is_api_gateway_event({'httpMethod': 'POST'}))

    def test_is_api_gateway_event_false(self):
        self.assertFalse(events._is_api_gateway_event({}))

    def test_is_storage_event_true(self):
        self.assertTrue(events._is_storage_event(S3_EVENT))

    def test_is_storage_event_false(self):
        self.assertFalse(events._is_storage_event({}))

    def test_parse_storage_event_s3(self):
        result = events._parse_storage_event(S3_EVENT)
        self.assertIsInstance(result, S3Event)

    def test_parse_storage_event_minio(self):
        result = events._parse_storage_event(MINIO_EVENT)
        self.assertIsInstance(result, MinioEvent)

    def test_parse_storage_event_onedata(self):
        result = events._parse_storage_event(ONEDATA_EVENT)
        self.assertIsInstance(result, OnedataEvent)

    def test_parse_storage_event_unknown(self):
        result = events._parse_storage_event(UNKNOWN_EVENT)
        self.assertIsNone(result)

    def test_parse_event_unknown(self):
        result = events.parse_event(UNKNOWN_EVENT)
        self.assertIsInstance(result, UnknownEvent)

    def test_parse_event_apigateway_wo_json_body(self):
        result = events.parse_event(APIGTW_EVENT_WO_JSON)
        self.assertIsInstance(result, ApiGatewayEvent)

    def test_parse_event_apigateway_w_json_body(self):
        result = events.parse_event(APIGTW_EVENT_W_JSON)
        self.assertIsInstance(result, S3Event)

    def test_parse_event_storage(self):
        result = events.parse_event(S3_EVENT)
        self.assertIsInstance(result, S3Event)


class ApiGatewayEventTest(unittest.TestCase):

    def test_set_event_params(self):
        api = ApiGatewayEvent(APIGTW_EVENT_W_JSON)
        self.assertEqual(api.body, S3_EVENT)
        api = ApiGatewayEvent(APIGTW_EVENT_WO_JSON)
        self.assertEqual(api.body, 'aXQgd29ya3Mh')

    def test_has_json_body(self):
        api = ApiGatewayEvent(APIGTW_EVENT_W_JSON)
        self.assertTrue(api.has_json_body())
        api = ApiGatewayEvent(APIGTW_EVENT_WO_JSON)
        self.assertFalse(api.has_json_body())

    def test_is_request_with_parameters(self):
        api = ApiGatewayEvent(APIGTW_EVENT_W_JSON)
        self.assertTrue(api._is_request_with_parameters())
        api = ApiGatewayEvent(APIGTW_EVENT_WO_JSON)
        self.assertFalse(api._is_request_with_parameters())

    def test_save_request_parameters(self):
        with mock.patch.dict('os.environ', {}, clear=True):
            ApiGatewayEvent(APIGTW_EVENT_W_JSON)
            self.assertEqual(os.environ, {"CONT_VAR_q1":"v1", "CONT_VAR_q2":"v2"})

    @mock.patch('faassupervisor.utils.FileUtils.create_file_with_content')
    @mock.patch('faassupervisor.utils.SysUtils.join_paths')
    def test_save_event_json(self, mock_join, mock_create):
        mock_join.return_value = '/tmp/test/file'
        api = ApiGatewayEvent(APIGTW_EVENT_W_JSON)
        result = api.save_event('/tmp/test')
        self.assertEqual(result, '/tmp/test/file')
        mock_create.assert_called_once_with('/tmp/test/file', S3_EVENT)

    @mock.patch('faassupervisor.utils.FileUtils.create_file_with_content')
    @mock.patch('faassupervisor.utils.SysUtils.join_paths')
    def test_save_event_binary(self, mock_join, mock_create):
        mock_join.return_value = '/tmp/test/file'
        api = ApiGatewayEvent(APIGTW_EVENT_WO_JSON)
        result = api.save_event('/tmp/test')
        self.assertEqual(result, '/tmp/test/file')
        mock_create.assert_called_once_with('/tmp/test/file', b'it works!', mode='wb')


class MinioEventTest(unittest.TestCase):

    def test_minio_event_creation(self):
        event = MinioEvent(MINIO_EVENT)
        self.assertEqual(event.object_key, "nature-wallpaper-229.jpg")
        self.assertEqual(event.bucket_arn, "arn:aws:s3:::images")
        self.assertEqual(event.bucket_name, "images")
        self.assertEqual(event.file_name, "nature-wallpaper-229.jpg")
        self.assertEqual(event.get_type(), "MINIO")


class OnedataEventTest(unittest.TestCase):

    def test_onedata_event_creation(self):
        event = OnedataEvent(ONEDATA_EVENT)
        self.assertEqual(event.object_key, "/my-onedata-space/files/file.txt")
        self.assertEqual(event.file_name, "file.txt")
        self.assertEqual(event.get_type(), "ONEDATA")


class S3EventTest(unittest.TestCase):

    def test_s3_event_creation(self):
        event = S3Event(S3_EVENT)
        self.assertEqual(event.object_key, "darknet-s3/input/dog.jpg")
        self.assertEqual(event.bucket_arn, "arn:aws:s3:::darknet-bucket")
        self.assertEqual(event.bucket_name, "darknet-bucket")
        self.assertEqual(event.file_name, "dog.jpg")
        self.assertEqual(event.get_type(), "S3")


class UnknownEventTest(unittest.TestCase):

    def test_set_event_params(self):
        event = UnknownEvent(UNKNOWN_EVENT)
        self.assertEqual(event.get_type(), "UNKNOWN")

    def test_set_event_params_empty(self):
        event = UnknownEvent("")
        self.assertEqual(event.get_type(), "UNKNOWN")

    @mock.patch('faassupervisor.utils.FileUtils.create_file_with_content')
    @mock.patch('faassupervisor.utils.SysUtils.join_paths')
    def test_save_event(self, mock_join, mock_create):
        mock_join.return_value = '/tmp/test/file'
        event = UnknownEvent(UNKNOWN_EVENT)
        event.save_event('/tmp/test')
        mock_create.assert_called_once_with('/tmp/test/file', UNKNOWN_EVENT)
